/*
 * CommLayer.java
 *
 * Created on 11 de Novembro de 2001, 20:34
 */

package br.shmem;
import java.net.*;
import java.io.*;
import java.security.*;

/** This is the communications layer.
 *
 * @author crolmos & tminc
 * @version 1.0
 */
public abstract class AbstractCommLayer {
    
    /** The algorithm used in all digital signatures in all hosts.
     */
    public static final String signatureAlgorithm = "MD2withRSA";
    public static final String keyPairGeneratorAlgorithm = "RSA";
    
    /** The <code>HostId</code> of this host.
     */
    protected HostId myId;
    
    /** the listening socket for this host.
     */
    private PrivateKey pKey;
    
    /** Creates new CommLayer */
    protected AbstractCommLayer() {
    }
    
    /** Constructs a new instance of the communications layer.
     * @param myId <code>HostId</code> of the local host.
     * @param pKey <code>PrivateKey</code> to be used for signing.
     * @param serverURL <code>URL</code> of the server.
     * @throws CommLayerException if any communication error occurs.
     */
    public AbstractCommLayer(HostId myId, PrivateKey pKey) {
        this.myId = myId;
        this.pKey = pKey;
        
    }
    
    public AbstractCommLayer(String hostname, int port)
    throws CommLayerException 
    {
        try {
            KeyPair keys = KeyPairGenerator.getInstance(keyPairGeneratorAlgorithm).generateKeyPair();
            pKey = keys.getPrivate();
            myId = new HostId(new URL("http://" + hostname + ":" + port), keys.getPublic());
        }
        catch (MalformedURLException e) {
            throw new CommLayerException(e.toString());
        }
        catch (NoSuchAlgorithmException e) {
            throw new CommLayerException(e.toString());
        }
    }
        
        /** connects with the <code>peer</code>.
         * @param peer remote host.
         * @throws CommLayerException if any communication error occurs.
         * @return an open connection to the <code>peer</code>
         */
        public Socket connect(String hostname, int port)
        throws CommLayerException {
            try {
                Socket s = new Socket(hostname, port);
                ObjectOutputStream out = new ObjectOutputStream(s.getOutputStream());
                out.writeObject(myId.getURL());
                out.flush();
                return s;
            }
            catch(IOException e) {
                throw new CommLayerException("Unreachable host: " + hostname + ":" + port);
            }
        }
  
              /** connects with the <code>peer</code>.
         * @param peer remote host.
         * @throws CommLayerException if any communication error occurs.
         * @return an open connection to the <code>peer</code>
         */
        public Socket connect(URL url)
        throws CommLayerException {
            return connect(url.getHost(), url.getPort());
        }
  
        /** Signs and send a message through the given socket. The message is signed using this host's private key.
         * @param s socket connecting the two hosts.
         * @param m <code>Message</code> to be sent.
         * @throws CommLayerException if any communication error occurs.
         * @throws InsecureObjectException if a digital signature cannot be created or verified
         */
        public void sendMessage(Socket s, Message m)
        throws CommLayerException, InsecureObjectException {
            try {
                SignedObject so = new SignedObject(m, pKey, Signature.getInstance(signatureAlgorithm));
                ObjectOutputStream out = new ObjectOutputStream(s.getOutputStream());
                out.writeObject(so);
                out.flush();
            }
            catch (IOException ioe) {
                throw new CommLayerException("Communication Error: " + ioe);
            }
            catch (NoSuchAlgorithmException nsae) {
                throw new CommLayerException("No such algorithm: " + signatureAlgorithm);
            }
            catch (InvalidKeyException ike) {
                throw new InsecureObjectException("Invalid key");
            }
            catch (SignatureException se) {
                throw new InsecureObjectException("Signature verification failed");
            }
        }
        
        /** Reads a message coming from the given socket and verifies its authenticity using the public key from <code>peer</code>.
         * @param s socket connecting the two hosts.
         * @param peer <code>HostId</code> of the peer.
         * @throws CommLayerException if any communication error occurs.
         * @throws InsecureObjectException if a digital signature cannot be created or verified
         * @return the received and verified <code>Message</code>.
         */
        public Message recvMessage(Socket s, HostId peer)
        throws CommLayerException, InsecureObjectException {
            try {
                ObjectInputStream in = new ObjectInputStream(s.getInputStream());
                SignedObject so = (SignedObject) in.readObject();
                if (!so.verify(peer.getKey(), Signature.getInstance(signatureAlgorithm)))
                    throw new InsecureObjectException("Signature verification failed");
                return (Message) so.getObject();
            }
            catch (IOException ioe) {
                throw new CommLayerException("Communication Error");
            }
            catch (NoSuchAlgorithmException nsae) {
                throw new CommLayerException("No such algorithm: " + signatureAlgorithm);
            }
            catch (InvalidKeyException ike) {
                throw new InsecureObjectException("Invalid key");
            }
            catch (ClassNotFoundException cnfe) {
                throw new CommLayerException("Impossible. Please notify us at tminc@linux.ime.usp.br");
            }
            catch (SignatureException se) {
                throw new InsecureObjectException("signature verification failed");
            }
        }
    }
