/*
 * SharedMemory.java
 *
 * Created on 4 de Dezembro de 2001, 20:32
 */

package br.shmem;

import br.shmem.client.*;
import br.shmem.util.*;

import java.util.*;
import java.io.Serializable;
import java.lang.reflect.*;

/** This class represents a connection to the global space in wich the objects are shared.
 * @author crolmos  & tminc
 */
public class SharedMemory {
    
    private ClientCommLayer commLayer;
    private boolean debug = false;
    private Hashtable objects;
    private Hashtable cache;
    private Hashtable locks;    
    
    /** Creates new SharedMemory */
    protected SharedMemory() {
    }
    
    /** Creates new SharedMemory
     * @param hostname the host where this object will be listening.
     * @param port the port in wich this object will be listening.
     */
    public SharedMemory(String hostname, int port) 
    throws CommLayerException
    {
        commLayer = new ClientCommLayer(this, hostname, port);
        objects = new Hashtable();
        locks = new Hashtable();
        cache = new Hashtable();
        commLayer.start();
    }
    
    /** Conects whith the specified server.
     * @param hostname the hostname of the server.
     * @param port the port used for login.
     */    
    public void connect(String hostname, int port) 
    throws CommLayerException, InvalidOperationException  
    {
        commLayer.login(hostname, port);
    }
    
    public boolean isDebugging() {
        return debug;
    }
    
    public void setDebugging(boolean flag) {
        debug = flag;
    }

    /** This method should not be called.
     */
    public void put(SharedObject obj) {
        objects.put(obj.getGlobalName(), obj);
    }

    /** Returns a list of all the objects shared by this <code>SharedMemory</code>.
     */    
    public List getInternalDescriptors() {
	List l = new ArrayList();
	Enumeration values = objects.elements();
	while (values.hasMoreElements()) {
	    SharedObject obj = (SharedObject) values.nextElement();
	    l.add(obj.getDescriptor());
	}
	return l;
    }

    /** Returns a list of all the objects cached by this <code>SharedMemory</code>.
     */    
    public List getExternalDescriptors() {
	List l = new ArrayList();
	Enumeration values = cache.elements();
	while (values.hasMoreElements()) {
	    SharedObject obj = (SharedObject) values.nextElement();
	    l.add(obj.getDescriptor());
	}
	return l;
    }
    
    /** This method should not be called.
     */
    public SharedObject getObject(String globalName) {
        return (SharedObject) objects.get(globalName);
    }
    
    /** This method should not be called.
     */
    protected void remove(String globalName) {
        objects.remove(globalName);
    }
    
    /** This method should not be called.
     */
    public SharedObject getCachedObject(String globalName) {
        return (SharedObject) cache.get(globalName);
    }
    
    private SharedReference getReference(SharedObject obj) {
        try {
        Class rc = Class.forName(obj.getReferenceClassName());
        Constructor c = rc.getConstructor(new Class[] {this.getClass(), String.class});
        return (SharedReference) c.newInstance(new Object[] {this, obj.getGlobalName()});
        }
        catch (ClassNotFoundException e) {
            return new SharedReference(this, obj.getGlobalName());
        }
        catch (NoSuchMethodException e) {
            return new SharedReference(this, obj.getGlobalName());
        }            
        catch (InstantiationException e) {
            return new SharedReference(this, obj.getGlobalName());
        }            
        catch (IllegalAccessException e) {
            return new SharedReference(this, obj.getGlobalName());
        }            
        catch (InvocationTargetException e) {
            return new SharedReference(this, obj.getGlobalName());
        }            
    }
    
    /** This method should not be called.
     */
    protected SharedReference getRemote(String globalName) 
    throws CommLayerException, InsecureObjectException, InvalidOperationException
    {
        SharedObject obj = commLayer.reload(globalName);
        obj.setLeaseTime(SharedObject.defaultLeaseTime);
        obj.setCanWrite(false);
        obj.setLastUpdated(System.currentTimeMillis());
        locks.put(globalName, new Mutex());
        cache.put(globalName, obj);
        return getReference(obj);
    }
    
    /** This method should not be called directly.
     */
    protected void lock(String globalName) {
        try {
	    ((Mutex) locks.get(globalName)).acquire();
        }
        catch (InterruptedException e) {
            throw new InterruptedError("interrupted while locking " + globalName);
        }
    }
    
    /** This method should not be called directly.
     */
    protected void unlock(String globalName) {
        ((Mutex) locks.get(globalName)).release();
    }
    
    /** Returns a <code>SharedReference</code> for a shared object.
     * @param globalName the global name of the object.
     * @return a <code>SharedReference</code> for this object.
     */    
    public SharedReference get(String globalName)
	throws CommLayerException, InsecureObjectException, InvalidOperationException    
    {
        SharedObject obj = getCachedObject(globalName);
        if (obj == null)
            return getRemote(globalName);
        else
            return getReference(obj);
    }
    
    /** Shares an object with the specified <code>globalName</code>.
     * @param globalName the global name of the shared object.
     * @param data the object to be shared.
     * @param referenceClassName the class of the <code>SharedReference</code> that should be returned when the method {@link #get(String) get} is called.
     * @return a shared reference for this object.
     */    
    public SharedReference share(String globalName, Serializable data, String referenceClassName)
	throws CommLayerException, InsecureObjectException, InvalidOperationException {
        SharedObject obj = new SharedObject(globalName, data, referenceClassName);
        commLayer.share(obj);
        obj.setCanWrite(false);
        objects.put(obj.getGlobalName(), obj);
        return get(globalName);
    }
    
    /** Shares an object with the specified <code>globalName</code>.
     * @param globalName the global name of the shared object.
     * @param data the object to be shared.
     * @return a shared reference for this object.
     */   
    public SharedReference share(String globalName, Serializable data)
	throws CommLayerException, InsecureObjectException, InvalidOperationException {
        SharedObject obj = new SharedObject(globalName, data, "br.shmem.SharedReference");
        commLayer.share(obj);
        obj.setCanWrite(false);
        objects.put(obj.getGlobalName(), obj);
        return get(globalName);
    }
    
    /** This method should not be called directly.
     */
    protected void unshare(SharedReference reference)
	throws CommLayerException, InsecureObjectException, InvalidOperationException {
        commLayer.unshare(reference.getGlobalName());
        objects.remove(reference.getGlobalName());
    }
    
    /** This method should not be called directly.
     */
    protected void validate(SharedReference reference)
	throws CommLayerException, InsecureObjectException, InvalidOperationException {
        String globalName = reference.getGlobalName();
        lock(globalName);
        try {
            SharedObject obj = getCachedObject(globalName);
            if (obj == null)
                throw new InvalidOperationException("Cannot validate: " + globalName + " is not cached.");
            HostId id = commLayer.validate(globalName, obj.getTimeStamp());
            if(id != null) {
                SharedObject newObj = commLayer.reload(globalName, id);
		newObj.setLeaseTime(obj.getLeaseTime());
                newObj.setCanWrite(obj.canWrite());
                newObj.setLastUpdated(System.currentTimeMillis());
                cache.put(globalName, newObj);
            }		
        }
        finally {
            unlock(globalName);
        }
    }
    
    /** This method should not be called directly.
     */
    protected void beginTrans(SharedReference reference)
    throws CommLayerException, InsecureObjectException, InvalidOperationException {
        String globalName = reference.getGlobalName();
        TimeStamp ts = commLayer.beginTrans(globalName);
        lock(globalName);
        try {
            SharedObject obj = getCachedObject(globalName);
            if (obj.getTimeStamp().before(ts)) {
                obj = commLayer.reload(globalName);
                cache.put(globalName, obj);
            }
            obj.setCanWrite(true);
        }
        finally {
            unlock(globalName);
        }
    }
    
    /** This method should not be called directly.
     */
    protected void commit(SharedReference reference)
    throws CommLayerException, InsecureObjectException, InvalidOperationException {
        lock(reference.getGlobalName());
        try {
            SharedObject obj = getCachedObject(reference.getGlobalName());
            HostId id = commLayer.commit(obj);
            obj.setCanWrite(false);
        }
        finally {
            unlock(reference.getGlobalName());
        }
    }
    
    /** This method should not be called directly.
     */
    protected void rollback(SharedReference reference)
    throws CommLayerException, InsecureObjectException, InvalidOperationException {
        String globalName = reference.getGlobalName();
        lock(globalName);
        try {
            HostId id = commLayer.rollback(globalName);
            SharedObject obj = commLayer.reload(globalName, id);
            obj.setCanWrite(false);
            cache.put(globalName, obj);
        }
        finally {
            unlock(globalName);
        }
    }

    /** Disconnects from the server.
     */
    public void disconnect()
    throws CommLayerException, InsecureObjectException, InvalidOperationException {
        commLayer.logout();
    }
    
}
