/*
 * ServerCommLayer.java
 *
 * Created on 24 de Novembro de 2001, 15:24
 */

package br.shmem.client;

import br.shmem.*;
import java.net.*;
import java.io.*;
import java.security.*;

/**
 *
 * @author  crolmos & tminc
 */
public class ClientCommLayer extends AbstractCommLayer {
    
    protected ServerSocket signedSocket;
    
    protected SharedMemory shmem;
    protected HostId serverId;
    private volatile long currentTime;
    
    /** Creates new ServerCommLayer */
    protected ClientCommLayer() {
    }
    
    public ClientCommLayer(SharedMemory shmem, String hostname, int port) 
    throws CommLayerException
    { 
        super(hostname, port);
        this.shmem = shmem;
    }

    public long getCurrentTime() {
	return currentTime;
    }
    
    
    protected void checkMessage(Message m, int type, String errorMsg)
    throws CommLayerException, InvalidOperationException {
        if (m.getType() != type) {
            if (shmem.isDebugging())
                System.out.println(errorMsg + "\nInvalid message type. Expected: "
                + type + " Received: " + m.getType());
            throw new CommLayerException(errorMsg + "\nInvalid message type. Expected: "
            + type + " Received: " + m.getType());
        }
        if (m.isError()) {
            if (shmem.isDebugging())
                System.out.println(m.getErrorDescription());
            throw new InvalidOperationException(m.getErrorDescription());
        }
    }
    
    public void start()
    throws CommLayerException {
        try {
            signedSocket = new ServerSocket(myId.getURL().getPort());
            
            SignedSocketListener ssl = new SignedSocketListener(this, signedSocket);
            new Thread(ssl).start();
        }
        catch(IOException ioe) {
            throw new CommLayerException("Can not open socket: " + ioe);
        }
    }
    
    public void login(String hostname, int port)
    throws CommLayerException, InvalidOperationException {
        try {
            Socket s = new Socket(hostname, port);
            ObjectOutputStream out = new ObjectOutputStream(s.getOutputStream());
            out.writeObject(Message.createLogin(myId));
            out.flush();
            ObjectInputStream in = new ObjectInputStream(s.getInputStream());
            Message m = (Message)in.readObject();
            checkMessage(m, Message.LOGIN, "Failed to login.");
            serverId = m.getHostId();
            if(shmem.isDebugging())
                System.out.println("Logged in: " + hostname + ":" + port + ".");
        }
        catch(IOException e) {
            throw new CommLayerException(e.toString());
        }
        catch(ClassNotFoundException e) {
            throw new CommLayerException(e.toString());
        }
    }
    
    /** Logs out from the server.
     * @throws CommLayerException if any communication error occurs.
     * @throws InsecureObjectException if a digital signature cannot be created or verified
     */
    public void logout()
    throws CommLayerException, InsecureObjectException, InvalidOperationException {
        Socket s = connect(serverId.getURL());
        sendMessage(s, Message.createLogout());
        Message m = recvMessage(s, serverId);
        checkMessage(m, Message.LOGOUT, "Failed to logout.");
        if(shmem.isDebugging())
            System.out.println("Logged out.");
    }
    
    /** Registers a global reference into the server's database.
     * @param obj the object to be shared.
     * @throws CommLayerException if any communication error occurs.
     * @throws InsecureObjectException if a digital signature cannot be created or verified
     * @throws InvalidOperationException if a remote error occurs, such as a global name not found.
     */
    public void share(SharedObject obj)
    throws CommLayerException, InsecureObjectException, InvalidOperationException {
        Socket s = connect(serverId.getURL());
        sendMessage(s, Message.createShare(obj.getGlobalName()));
        Message m = recvMessage(s, serverId);
        checkMessage(m, Message.SHARE, "Failed to share: " + obj.getGlobalName() + ".");
        TimeStamp stamp = m.getTimeStamp();
        obj.setTimeStamp(stamp);
	if(shmem.isDebugging())
            System.out.println("Shared: " + obj.getGlobalName() + " (" + stamp + ").");
    }
    
    /** removes a reference from the global space.
     * @param globalName the object identifier in the global space
     * @throws CommLayerException if any communication error occurs.
     * @throws InsecureObjectException if a digital signature cannot be created or verified
     * @throws InvalidOperationException if a remote error occurs, such as a global name not found.
     */
    public void unshare(String globalName)
    throws CommLayerException, InsecureObjectException, InvalidOperationException {
        Socket s = connect(serverId.getURL());
        sendMessage(s, Message.createUnshare(globalName));
        Message m = recvMessage(s, serverId);
        checkMessage(m, Message.UNSHARE, "Failed to unshare: " + globalName + ".");
        if(shmem.isDebugging())
            System.out.println("Unshared: " + globalName + ".");
    }
    
    /** Asks if the local copy of the remote object is still valid. If the copy is too old, the <code>HostId</code> returned by the method can be used to contact the owner.
     * @param globalName the object identifier in the global space
     * @param stamp
     * @throws CommLayerException if any communication error occurs.
     * @throws InsecureObjectException if a digital signature cannot be created or verified
     * @throws InvalidOperationException if a remote error occurs, such as a global name not found.
     * @return
     */
    public HostId validate(String globalName, TimeStamp stamp)
    throws CommLayerException, InsecureObjectException, InvalidOperationException {
        Socket s = connect(serverId.getURL());
        sendMessage(s, Message.createValidate(globalName, stamp));
        Message m = recvMessage(s, serverId);
        checkMessage(m, Message.VALIDATE, "Failed to validate: " + globalName + ".");
        HostId id = m.getHostId();
        if(shmem.isDebugging()) {
            System.out.print("Validated: " + globalName);
            if(id != null)
                System.out.println(" Need to be updated from: " + id);
            else
                System.out.println(" Object is up to date.");
        }
        return id;
    }
    
    /** Asks the server for the owner of the given global name.
     * @param globalName the object identifier in the global space
     * @throws CommLayerException if any communication error occurs.
     * @throws InsecureObjectException if a digital signature cannot be created or verified
     * @throws InvalidOperationException if a remote error occurs, such as a global name not found.
     * @return
     */
    public HostId request(String globalName)
    throws CommLayerException, InsecureObjectException, InvalidOperationException {
        Socket s = connect(serverId.getURL());
        sendMessage(s, Message.createRequest(globalName));
        Message m = recvMessage(s, serverId);
        checkMessage(m, Message.REQUEST, "Failed to request: " + globalName + ".");
        HostId id = m.getHostId();
        if(shmem.isDebugging())
            System.out.println("Requested: " + globalName + " Owner: " + id);
        return id;
    }
    
    /** Try to initiate a remote writing transaction.
     * @param globalName the object identifier in the global space
     * @throws CommLayerException if any communication error occurs.
     * @throws InsecureObjectException if a digital signature cannot be created or verified
     * @throws InvalidOperationException if a remote error occurs, such as a global name not found.
     * @return
     */
    public TimeStamp beginTrans(String globalName)
    throws CommLayerException, InsecureObjectException, InvalidOperationException {
        Socket s = connect(serverId.getURL());
        sendMessage(s, Message.createBeginTrans(globalName));
        Message m = recvMessage(s, serverId);
        checkMessage(m, Message.BEGIN_TRANS, "Failed to begin transaction : " + globalName + ".");
        TimeStamp stamp = m.getTimeStamp();
        if(shmem.isDebugging())
            System.out.println("Transaction: " + globalName + " Actual TimeStamp: " + stamp);
        return stamp;
        
    }
    
    /** Begins the commit process. Updates the <code>TimeStamp</code> of the shared object and indicates where to send it.
     * @param globalName the object identifier in the global space
     * @param obj
     * @throws CommLayerException if any communication error occurs.
     * @throws InsecureObjectException if a digital signature cannot be created or verified
     * @throws InvalidOperationException if a remote error occurs, such as a global name not found.
     * @return
     */
    public HostId commit(SharedObject obj)
    throws CommLayerException, InsecureObjectException, InvalidOperationException {
        Socket s = connect(serverId.getURL());
        sendMessage(s, Message.createCommit(obj.getGlobalName()));
        Message m = recvMessage(s, serverId);
        checkMessage(m, Message.COMMIT, "Failed to commit: " + obj.getGlobalName() + ".");
        HostId id = m.getHostId();
        obj.setTimeStamp(m.getTimeStamp());
	currentTime = m.getTimeStamp().getDate().getTime();
        if(shmem.isDebugging())
            System.out.println("Commited: " + obj.getGlobalName() + " Owner: " + id);
        return id;
    }
    
    /** Releases the writing lock, abandoning all changes and indicates where to get the most recent copy of the shared object.
     * @param globalName the object identifier in the global space
     * @throws CommLayerException if any communication error occurs.
     * @throws InsecureObjectException if a digital signature cannot be created or verified
     * @throws InvalidOperationException if a remote error occurs, such as a global name not found.
     * @return
     */
    public HostId rollback(String globalName)
    throws CommLayerException, InsecureObjectException, InvalidOperationException {
        Socket s = connect(serverId.getURL());
        sendMessage(s, Message.createRollback(globalName));
        Message m = recvMessage(s, serverId);
        checkMessage(m, Message.ROLLBACK, "Failed to rollback: " + globalName + ".");
        HostId id = m.getHostId();
        if(shmem.isDebugging())
            System.out.println("Rolled back: " + globalName + " Owner: " + id);
        return id;
    }
    
    /** Retrieves the shared object data identified by the global name.
     * @param obj
     * @param owner
     * @throws CommLayerException if any communication error occurs.
     * @throws InsecureObjectException if a digital signature cannot be created or verified
     * @throws InvalidOperationException if a remote error occurs, such as a global name not found.
     * @return
     */
    public SharedObject get(String globalName, HostId owner)
    throws CommLayerException, InsecureObjectException, InvalidOperationException {
        Socket s = connect(owner.getURL());
        sendMessage(s, Message.createGet(globalName, myId));
        Message m = recvMessage(s, owner);
        checkMessage(m, Message.GET, "Failed to get: " + globalName + " from " + owner + ".");
        return m.getData();
    }
    
    /** Retrieves the shared object data identified by the global name.
     * @param obj
     * @param owner
     * @throws CommLayerException if any communication error occurs.
     * @throws InsecureObjectException if a digital signature cannot be created or verified
     * @throws InvalidOperationException if a remote error occurs, such as a global name not found.
     * @return
     */
    public SharedObject reload(String globalName, HostId owner)
    throws CommLayerException, InsecureObjectException, InvalidOperationException {
        Socket s = connect(owner.getURL());
        sendMessage(s, Message.createReload(globalName, myId));
        Message m = recvMessage(s, owner);
        checkMessage(m, Message.RELOAD, "Failed to reload: " + globalName + " from " + owner + ".");
        return m.getData();
    }
    
    public SharedObject reload(String globalName) 
    throws CommLayerException, InsecureObjectException, InvalidOperationException {
        return reload(globalName, request(globalName));
    }
    
    public void sendCached(Socket s, String globalName)
    throws CommLayerException, InsecureObjectException, InvalidOperationException {
        SharedObject obj = null;
        try {
            obj = shmem.getCachedObject(globalName);
        }
        catch (NullPointerException e) {
            sendMessage(s, Message.createError(Message.GET, "Object: " + globalName + " not found."));
            if (shmem.isDebugging())
                System.out.println("Object: " + globalName + " not found.");
            return;
        }
        sendMessage(s, Message.createOK(Message.GET, globalName, obj));
        if (shmem.isDebugging())
            System.out.println("Object: " + globalName + " sent.");
    }
    
    public void send(Socket s, String globalName)
    throws CommLayerException, InsecureObjectException, InvalidOperationException {
        SharedObject obj = null;
        try {
            obj = shmem.getObject(globalName);
        }
        catch (NullPointerException e) {
            sendMessage(s, Message.createError(Message.RELOAD, "Object: " + globalName + " not found."));
            if (shmem.isDebugging())
                System.out.println("Object: " + globalName + " not found.");
            return;
        }
        sendMessage(s, Message.createOK(Message.RELOAD, globalName, obj));
        if (shmem.isDebugging())
            System.out.println("Object: " + globalName + " sent(reloaded).");
    }
    
    
    /** Message sent from the owner to the server to finish the commit process.
     * @param globalName the object identifier in the global space
     * @param stamp
     * @throws CommLayerException if any communication error occurs.
     * @throws InsecureObjectException if a digital signature cannot be created or verified
     * @throws InvalidOperationException if a remote error occurs, such as a global name not found.
     */
    public void unlock(String globalName, HostId writer, TimeStamp stamp)
    throws CommLayerException, InsecureObjectException, InvalidOperationException {
        Socket s = connect(serverId.getURL());
        sendMessage(s, Message.createUnlock(globalName, writer, stamp));
        Message m = recvMessage(s, serverId);
        checkMessage(m, Message.UNLOCK, "Failed to unlock: " + globalName + ".");
        if(shmem.isDebugging())
            System.out.println("Unlocked: " + globalName + ".");
    }
    
    public HostId identify(URL remoteURL)
    throws CommLayerException, InsecureObjectException, InvalidOperationException {
        if (remoteURL.equals(serverId.getURL()))
            return serverId;
        Socket s = connect(serverId.getURL());
        sendMessage(s, Message.createIdentify(remoteURL));
        Message m = recvMessage(s, serverId);
        checkMessage(m, Message.IDENTIFY, "Failed to identify: " + remoteURL + ".");
        HostId id = m.getHostId();
        if(shmem.isDebugging())
            System.out.println("Identified: " + remoteURL + ".");
        return id;
    }
    
    public void update(String globalName, HostId writer)
    throws CommLayerException, InsecureObjectException, InvalidOperationException {
        SharedObject obj = get(globalName, writer);
        unlock(globalName, writer, obj.getTimeStamp());
        shmem.put(obj);
    }
    
    class SignedSocketListener implements Runnable {
        ServerSocket serverSocket;
        ClientCommLayer commLayer;
        
        public SignedSocketListener(ClientCommLayer commLayer, ServerSocket serverSocket) {
            this.commLayer = commLayer;
            this.serverSocket = serverSocket;
        }
        
        public void run() {
            for (;;) {
                Socket s = null;
                try {
                    s = serverSocket.accept();
                }
                catch (IOException e) {
                    System.err.println(e);
                }
                
                SignedConnection connection = new SignedConnection(commLayer, s);
                new Thread(connection).start();
            }
        }
    }
}
