/*
 * ServerCommLayer.java
 *
 * Created on 24 de Novembro de 2001, 15:24
 */

package br.shmem.server;

import br.shmem.*;
import java.net.*;
import java.io.*;
import java.security.*;

/**
 *
 * @author  crolmos & tminc
 */
public class ServerCommLayer extends AbstractCommLayer {
    
    protected ServerSocket plainSocket;
    protected ServerSocket signedSocket;
    
    protected Server server;
    
    /** Creates new ServerCommLayer */
    protected ServerCommLayer() {
    }
    
    public ServerCommLayer(Server server)
    throws CommLayerException {
        super(server.getHostId(), server.getPrivateKey());
        this.server = server;
    }
    
    public void start()
    throws CommLayerException {
        try {
            plainSocket = new ServerSocket(server.getPlainSocketURL().getPort());
            signedSocket = new ServerSocket(myId.getURL().getPort());
            
            PlainSocketListener psl = new PlainSocketListener(server, plainSocket);
            new Thread(psl).start();
            
            SignedSocketListener ssl = new SignedSocketListener(server, signedSocket);
            new Thread(ssl).start();
        }
        catch(IOException ioe) {
            throw new CommLayerException("Can not open socket: " + ioe);
        }
    }
    
    public Server getServer() {
        return server;
    }
    
    public void login(Socket s, HostId client)
    throws CommLayerException {
        try {
            ObjectOutputStream out = new ObjectOutputStream(s.getOutputStream());
            
            if(server.login(client)) {
                out.writeObject(Message.createLogin(myId));
                if(server.isDebugging())
                    System.out.println("Logged in: " + client + ".");
            }
            else {
                out.writeObject(Message.createError(Message.LOGIN, "Already logged in: " + client.getURL()));
                if(server.isDebugging())
                    System.out.println(client + " tried to login but was alredy logged.");
            }
            out.flush();
        }
        catch(IOException e) {
            throw new CommLayerException(e.toString());
        }
    }
    
    public void logout(Socket s, URL client)
    throws CommLayerException, InsecureObjectException {
        if(server.logout(client)) {
            sendMessage(s, Message.createLogout());
            if(server.isDebugging())
                System.out.println("Logged out: " + client + ".");
        }
        else {
            sendMessage(s, Message.createError(Message.LOGOUT, "Not logged in: " + client));
            if(server.isDebugging())
                System.out.println(client + " tried to logout but was not logged.");
        }
    }
    
    public boolean ping(URL client) {
        try {
            Socket s = connect(client);
            sendMessage(s, Message.createPing());
            s.close();
        }
        catch(IOException e) {
            if(server.isDebugging())
                System.out.println("Ping failed for: " + client + "\n" + e.toString());
            return false;
        }
        catch(CommLayerException e) {
            if(server.isDebugging())
                System.out.println("Ping failed for: " + client + "\n" + e.toString());
            return false;
        }
        catch(InsecureObjectException e) {
            if(server.isDebugging())
                System.out.println("Ping failed for: " + client + "\n" + e.toString());
            return false;
        }
        if(server.isDebugging())
            System.out.println("Ping: " + client + ".");
        return true;
    }
    
    public void share(Socket s, URL client, String globalName)
    throws CommLayerException, InsecureObjectException{
        try {
            TimeStamp stamp = server.share(client, globalName, true);
            sendMessage(s, Message.createOK(Message.SHARE, stamp));
            if(server.isDebugging())
                System.out.println("Shared : " + globalName + " for: " + client + ".");
        }
        catch (AlreadySharedException e) {
            sendMessage(s, Message.createError(Message.SHARE, e.getMessage()));
            if(server.isDebugging())
                System.out.println("Couldnt share: " + globalName + " for: " + client + ", already shared.");
        }
    }
    
    protected void unshare(Socket s, URL client, String globalName)
    throws CommLayerException, InsecureObjectException{
        try {
            server.unshare(client, globalName);
            sendMessage(s, Message.createOK(Message.UNSHARE));
            if(server.isDebugging())
                System.out.println(client + " unshared: " + globalName + ".");
        }
        catch (NotSharedException e) {
            sendMessage(s, Message.createError(Message.UNSHARE, e.getMessage()));
            if(server.isDebugging())
                System.out.println("Couldnt unshare: " + globalName + " for: " + client + ", unshared.");
        }
        catch (InvalidOperationException e) {
            sendMessage(s, Message.createError(Message.UNSHARE, e.getMessage()));
            if(server.isDebugging())
                System.out.println("Couldnt unshare: " + globalName + " for: " + client + ", invalid owner.");
        }
    }
    
    protected void request(Socket s, String globalName)
    throws CommLayerException, InsecureObjectException{
        try {
            HostId id = server.request(globalName);
            sendMessage(s, Message.createOK(Message.REQUEST, id));
            if(server.isDebugging())
                System.out.println("Request for: " + globalName + ".");
        }
        catch (NotSharedException e) {
            sendMessage(s, Message.createError(Message.REQUEST, e.getMessage()));
            if(server.isDebugging())
                System.out.println("Couldnt response request for: " + globalName + ", not shared.");
        }
    }
    
    protected void validate(Socket s, String globalName, TimeStamp stamp)
    throws CommLayerException, InsecureObjectException{
        try {
            HostId id = server.validate(globalName, stamp);
            sendMessage(s, Message.createOK(Message.VALIDATE, id));
            if(server.isDebugging())
                System.out.println("Validate (" + globalName + "): " + id);
        }
        catch (NotSharedException e) {
            sendMessage(s, Message.createError(Message.VALIDATE, e.getMessage()));
            if(server.isDebugging())
                System.out.println("Couldnt validate: " + globalName + ", not shared.");
        }
        catch (InsecureObjectException e) {
            sendMessage(s, Message.createError(Message.VALIDATE, e.getMessage()));
            if(server.isDebugging())
                System.out.println("Couldnt validate: " + globalName + ", insecure object.");
        }
    }
    
    protected void beginTrans(Socket s, URL client, String globalName)
    throws CommLayerException, InsecureObjectException{
        try {
            TimeStamp stamp = server.beginTrans(client, globalName);
            HostId id = server.request(globalName);
            sendMessage(s, Message.createOK(Message.BEGIN_TRANS, id, stamp));
            if(server.isDebugging())
                System.out.println("Initialized transaction for: " + globalName + " to: " + client);
        }
        catch (NotSharedException e) {
            sendMessage(s, Message.createError(Message.BEGIN_TRANS, e.getMessage()));
            if(server.isDebugging())
                System.out.println("Couldnt initialize transaction for: " + globalName + " to: " + client + ", not shared.");
        }
        catch (AlreadyLockedException e) {
            sendMessage(s, Message.createError(Message.BEGIN_TRANS, e.getMessage()));
            if(server.isDebugging())
                System.out.println("Couldnt initialize transaction for: " + globalName + " to: " + client + ", already locked.");
        }
        catch (ObjectNotLockableException e) {
            sendMessage(s, Message.createError(Message.BEGIN_TRANS, e.getMessage()));
            if(server.isDebugging())
                System.out.println("Couldnt initialize transaction for: " + globalName + " to: " + client + ", read only.");
        }
    }
    
    protected void commit(Socket s, URL client, String globalName)
    throws CommLayerException, InsecureObjectException{
        try {
            TimeStamp stamp = server.commit(client, globalName);
            HostId id = server.request(globalName);
            HostId clientId = server.getHostId(client);
            sendMessage(s, Message.createOK(Message.COMMIT, id, stamp));
            update(globalName, clientId);
            if(server.isDebugging())
                System.out.println("Commited transaction for: " + globalName + " to: " + client + ".");
        }
        catch (NotSharedException e) {
            sendMessage(s, Message.createError(Message.COMMIT, e.getMessage()));
            if(server.isDebugging())
                System.out.println("Couldnt commit transaction for: " + globalName + " to: " + client + ", not shared.");
        }
        catch (InvalidOperationException e) {
            sendMessage(s, Message.createError(Message.COMMIT, e.getMessage()));
            if(server.isDebugging())
                System.out.println("Couldnt commit transaction for: " + globalName + " to: " + client + ", not the owner of the lock.");
        }
        catch (HostNotFoundException e) {
            sendMessage(s, Message.createError(Message.COMMIT, e.getMessage()));
            if(server.isDebugging())
                System.out.println("Couldnt commit transaction for: " + globalName + " to: " + client + ", writer not found.");
        }
    }
    
    protected void rollback(Socket s, URL client, String globalName)
    throws CommLayerException, InsecureObjectException{
        try {
            server.rollback(client, globalName);
	    HostId owner = server.request(globalName);
            sendMessage(s, Message.createOK(Message.ROLLBACK, owner));
            if(server.isDebugging())
                System.out.println("Rolled back transaction for: " + globalName + " to: " + client + ", not shared.");
        }
        catch (NotSharedException e) {
            sendMessage(s, Message.createError(Message.ROLLBACK, e.getMessage()));
            if(server.isDebugging())
                System.out.println("Couldnt rollback transaction for: " + globalName + " to: " + client + ", not shared.");
        }
        catch (NotLockedException e) {
            sendMessage(s, Message.createError(Message.ROLLBACK, e.getMessage()));
            if(server.isDebugging())
                System.out.println("Couldnt rollback transaction for: " + globalName + " to: " + client + ", not locked.");
        }
    }
    
    protected void unlock(Socket s, URL client, String globalName, TimeStamp stamp)
    throws CommLayerException, InsecureObjectException{
        try {
            try {
                server.unlock(client, globalName, stamp);
            }
            catch (InsecureObjectException e) {
                sendMessage(s, Message.createError(Message.UNLOCK, e.getMessage()));
                if(server.isDebugging())
                    System.out.println("Couldnt unlock: " + globalName + " to: " + client + ", insecure object.");
                return;
            }
            sendMessage(s, Message.createOK(Message.UNLOCK));
            if(server.isDebugging())
                System.out.println("Unlocked object: " + globalName + " to: " + client + ".");
        }
        catch (NotSharedException e) {
            sendMessage(s, Message.createError(Message.UNLOCK, e.getMessage()));
            if(server.isDebugging())
                System.out.println("Couldnt unlock: " + globalName + " to: " + client + ", not shared.");
        }
        catch (NotLockedException e) {
            sendMessage(s, Message.createError(Message.UNLOCK, e.getMessage()));
            if(server.isDebugging())
                System.out.println("Couldnt unlock: " + globalName + " to: " + client + ", not locked.");
        }
        catch (InsecureObjectException e) {
            sendMessage(s, Message.createError(Message.UNLOCK, e.getMessage()));
            if(server.isDebugging())
                System.out.println("Couldnt unlock: " + globalName + " to: " + client + ", insecure object.");
        }
    }
    
    protected void identify(Socket s, URL url)
    throws CommLayerException, InsecureObjectException{
        try {
            HostId id = server.getHostId(url);
            sendMessage(s, Message.createOK(Message.IDENTIFY, id));
            if(server.isDebugging())
                System.out.println("Identify for: " + url + ".");
        }
        catch (HostNotFoundException e) {
            sendMessage(s, Message.createError(Message.IDENTIFY, e.getMessage()));
            if(server.isDebugging())
                System.out.println("Couldnt respond identify for: " + url + ", host not found.");
        }
    }
    
    protected void update(String globalName, HostId writer)
    throws CommLayerException, InsecureObjectException, NotSharedException {
        try {
            URL owner = server.request(globalName).getURL();
            Socket s = connect(owner);
            sendMessage(s, Message.createUpdate(globalName, writer));
            s.close();
        }
        catch(IOException e) {
            if(server.isDebugging())
                System.out.println("Update failed for: " + globalName + "\n" + e.toString());
            throw new CommLayerException("Update failed for: " + globalName + "\n" + e.toString());
        }
        if(server.isDebugging())
            System.out.println("Update: " + globalName + ".");
    }
    
    class PlainSocketListener implements Runnable {
        ServerSocket serverSocket;
        Server server;
        
        public PlainSocketListener(Server server, ServerSocket serverSocket) {
            this.server = server;
            this.serverSocket = serverSocket;
        }
        
        public void run() {
            for (;;) {
                Socket clientSocket = null;
                try {
                    clientSocket = serverSocket.accept();
                }
                catch (IOException e) {
                    System.err.println(e);
                }
                
                PlainConnection connection = new PlainConnection(server, clientSocket);
                new Thread(connection).start();
            }
        }
    }
    
    class SignedSocketListener implements Runnable {
        ServerSocket serverSocket;
        Server server;
        
        public SignedSocketListener(Server server, ServerSocket serverSocket) {
            this.server = server;
            this.serverSocket = serverSocket;
        }
        
        public void run() {
            for (;;) {
                Socket clientSocket = null;
                try {
                    clientSocket = serverSocket.accept();
                    SignedConnection connection = new SignedConnection(server, clientSocket);
                    new Thread(connection).start();
                }
                catch (IOException e) {
                }
            }
        }
    }
}
